# See also: https://developer.microsoft.com/en-us/graph/docs/api-reference/v1.0/resources/excel
# TODO: proper exception handling in xlwings base package so it can be used here
import sys
import logging

from werkzeug.routing import PathConverter

import xlwings as xw
from xlwings.rest.serializers import (serialize_book, serialize_sheet, serialize_app, serialize_chart, serialize_name,
                                      serialize_picture, serialize_range, serialize_shape)

try:
    import flask
    from flask import Flask, jsonify, request, abort
except ImportError:
    raise Exception("To use the xlwings REST API server, you need Flask>=1.0.0 installed.")


api = Flask(__name__)
logger = logging.getLogger(__name__)


class EverythingConverter(PathConverter):
    regex = '.*?'


if sys.platform.startswith('darwin'):
    # Hack to allow leading slashes on Mac
    api.url_map.converters['path'] = EverythingConverter


def get_book_object(fullname=None, name_or_ix=None, app_ix=None):
    assert fullname is None or name_or_ix is None
    if fullname:
        try:
            return xw.Book(fullname)
        except Exception as e:
            logger.exception(str(e))
            abort(500, str(e))
    elif name_or_ix:
        if name_or_ix.isdigit():
            name_or_ix = int(name_or_ix)
        app = xw.apps[int(app_ix)] if app_ix else xw.apps.active
        try:
            return app.books[name_or_ix]
        except KeyError as e:
            logger.exception(str(e))
            abort(500, "Couldn't find Book: " + str(e))
        except Exception as e:
            logger.exception(str(e))
            abort(500, str(e))


def get_sheet_object(book, name_or_id):
    if name_or_id.isdigit():
        name_or_id = int(name_or_id)
    return book.sheets[name_or_id]


@api.route('/apps', methods=['GET'])
def get_apps():
    return jsonify(apps=[serialize_app(app)
                         for app in xw.apps])


@api.route('/apps/<pid>', methods=['GET'])
def get_app(pid):
    return jsonify(serialize_app(xw.apps[int(pid)]))


@api.route('/apps/<pid>/books', methods=['GET'])
@api.route('/books', methods=['GET'])
def get_books(pid=None):
    books = xw.apps[int(pid)].books if pid else xw.books
    return jsonify(books=[serialize_book(book)
                          for book in books])


@api.route('/apps/<pid>/books/<book_name_or_ix>', methods=['GET'])
@api.route('/books/<book_name_or_ix>', methods=['GET'])
@api.route('/book/<path:fullname_or_name>', methods=['GET'])
def get_book(book_name_or_ix=None, fullname_or_name=None, pid=None):
    book = get_book_object(fullname_or_name, book_name_or_ix, pid)
    return jsonify(serialize_book(book))


@api.route('/apps/<pid>/books/<book_name_or_ix>/sheets', methods=['GET'])
@api.route('/books/<book_name_or_ix>/sheets', methods=['GET'])
@api.route('/book/<path:fullname_or_name>/sheets', methods=['GET'])
def get_sheets(book_name_or_ix=None, fullname_or_name=None, pid=None):
    book = get_book_object(fullname_or_name, book_name_or_ix, pid)
    return jsonify(sheets=[serialize_sheet(sheet)
                           for sheet in book.sheets])


@api.route('/apps/<pid>/books/<book_name_or_ix>/sheets/<sheet_name_or_ix>', methods=['GET'])
@api.route('/books/<book_name_or_ix>/sheets/<sheet_name_or_ix>', methods=['GET'])
@api.route('/book/<path:fullname_or_name>/sheets/<sheet_name_or_ix>', methods=['GET'])
def get_sheet(sheet_name_or_ix, book_name_or_ix=None, fullname_or_name=None, pid=None):
    book = get_book_object(fullname_or_name, book_name_or_ix, pid)
    sheet = get_sheet_object(book, sheet_name_or_ix)
    return jsonify(serialize_sheet(sheet))


@api.route('/apps/<pid>/books/<book_name_or_ix>/sheets/<sheet_name_or_ix>/range', methods=['GET'])
@api.route('/books/<book_name_or_ix>/sheets/<sheet_name_or_ix>/range', methods=['GET'])
@api.route('/book/<path:fullname_or_name>/sheets/<sheet_name_or_ix>/range', methods=['GET'])
def get_range(sheet_name_or_ix, book_name_or_ix=None, fullname_or_name=None, pid=None):
    book = get_book_object(fullname_or_name, book_name_or_ix, pid)
    sheet = get_sheet_object(book, sheet_name_or_ix)
    return jsonify(serialize_range(sheet.used_range))


@api.route('/apps/<pid>/books/<book_name_or_ix>/sheets/<sheet_name_or_ix>/range/<address>', methods=['GET'])
@api.route('/books/<book_name_or_ix>/sheets/<sheet_name_or_ix>/range/<address>', methods=['GET'])
@api.route('/book/<path:fullname_or_name>/sheets/<sheet_name_or_ix>/range/<address>', methods=['GET'])
def get_range_address(address, sheet_name_or_ix, book_name_or_ix=None, fullname_or_name=None, pid=None):
    book = get_book_object(fullname_or_name, book_name_or_ix, pid)
    sheet = get_sheet_object(book, sheet_name_or_ix)
    options = {k: v for k, v in request.args.items()}
    return jsonify(serialize_range(sheet.range(address).options(**options)))


@api.route('/apps/<pid>/books/<book_name_or_ix>/names', methods=['GET'])
@api.route('/books/<book_name_or_ix>/names', methods=['GET'])
@api.route('/book/<path:fullname_or_name>/names', methods=['GET'])
def get_book_names(book_name_or_ix=None, fullname_or_name=None, pid=None):
    book = get_book_object(fullname_or_name, book_name_or_ix, pid)
    return jsonify(names=[serialize_name(name) for name in book.names])


@api.route('/apps/<pid>/books/<book_name_or_ix>/names/<name>', methods=['GET'])
@api.route('/books/<book_name_or_ix>/names/<name>', methods=['GET'])
@api.route('/book/<path:fullname_or_name>/names/<name>', methods=['GET'])
def get_book_name(name, book_name_or_ix=None, fullname_or_name=None, pid=None):
    book = get_book_object(fullname_or_name, book_name_or_ix, pid)
    return jsonify(serialize_name(book.names[name]))


@api.route('/apps/<pid>/books/<book_name_or_ix>/names/<name>/range', methods=['GET'])
@api.route('/books/<book_name_or_ix>/names/<name>/range', methods=['GET'])
@api.route('/book/<path:fullname_or_name>/names/<name>/range', methods=['GET'])
def get_book_name_range(name, book_name_or_ix=None, fullname_or_name=None, pid=None):
    book = get_book_object(fullname_or_name, book_name_or_ix, pid)
    return jsonify(serialize_range(book.names[name].refers_to_range))


@api.route('/apps/<pid>/books/<book_name_or_ix>/sheets/<sheet_name_or_ix>/names', methods=['GET'])
@api.route('/books/<book_name_or_ix>/sheets/<sheet_name_or_ix>/names', methods=['GET'])
@api.route('/book/<path:fullname_or_name>/sheets/<sheet_name_or_ix>/names', methods=['GET'])
def get_sheet_names(sheet_name_or_ix, book_name_or_ix=None, fullname_or_name=None, pid=None):
    book = get_book_object(fullname_or_name, book_name_or_ix, pid)
    sheet = get_sheet_object(book, sheet_name_or_ix)
    return jsonify(names=[serialize_name(name) for name in sheet.names])


@api.route('/apps/<pid>/books/<book_name_or_ix>/sheets/<sheet_name_or_ix>/names/<sheet_scope_name>', methods=['GET'])
@api.route('/books/<book_name_or_ix>/sheets/<sheet_name_or_ix>/names/<sheet_scope_name>', methods=['GET'])
@api.route('/book/<path:fullname_or_name>/sheets/<sheet_name_or_ix>/names/<sheet_scope_name>', methods=['GET'])
def get_sheet_name(sheet_name_or_ix, sheet_scope_name, book_name_or_ix=None, fullname_or_name=None, pid=None):
    book = get_book_object(fullname_or_name, book_name_or_ix, pid)
    sheet = get_sheet_object(book, sheet_name_or_ix)
    return jsonify(serialize_name(sheet.names[sheet_scope_name]))


@api.route('/apps/<pid>/books/<book_name_or_ix>/sheets/<sheet_name_or_ix>/names/<sheet_scope_name>/range', methods=['GET'])
@api.route('/books/<book_name_or_ix>/sheets/<sheet_name_or_ix>/names/<sheet_scope_name>/range', methods=['GET'])
@api.route('/book/<path:fullname_or_name>/sheets/<sheet_name_or_ix>/names/<sheet_scope_name>/range', methods=['GET'])
def get_sheet_name_range(sheet_name_or_ix, sheet_scope_name, book_name_or_ix=None, fullname_or_name=None, pid=None):
    book = get_book_object(fullname_or_name, book_name_or_ix, pid)
    sheet = get_sheet_object(book, sheet_name_or_ix)
    return jsonify(serialize_range(sheet.names[sheet_scope_name].refers_to_range))


@api.route('/apps/<pid>/books/<book_name_or_ix>/sheets/<sheet_name_or_ix>/charts', methods=['GET'])
@api.route('/books/<book_name_or_ix>/sheets/<sheet_name_or_ix>/charts', methods=['GET'])
@api.route('/book/<path:fullname_or_name>/sheets/<sheet_name_or_ix>/charts', methods=['GET'])
def get_charts(sheet_name_or_ix, book_name_or_ix=None, fullname_or_name=None, pid=None):
    book = get_book_object(fullname_or_name, book_name_or_ix, pid)
    sheet = get_sheet_object(book, sheet_name_or_ix)
    return jsonify(charts=[serialize_chart(chart) for chart in sheet.charts])


@api.route('/apps/<pid>/books/<book_name_or_ix>/sheets/<sheet_name_or_ix>/charts/<chart_name_or_ix>', methods=['GET'])
@api.route('/books/<book_name_or_ix>/sheets/<sheet_name_or_ix>/charts/<chart_name_or_ix>', methods=['GET'])
@api.route('/book/<path:fullname_or_name>/sheets/<sheet_name_or_ix>/charts/<chart_name_or_ix>', methods=['GET'])
def get_chart(sheet_name_or_ix, chart_name_or_ix, book_name_or_ix=None, fullname_or_name=None, pid=None):
    book = get_book_object(fullname_or_name, book_name_or_ix, pid)
    sheet = get_sheet_object(book, sheet_name_or_ix)
    chart = int(chart_name_or_ix) if chart_name_or_ix.isdigit() else chart_name_or_ix
    return jsonify(serialize_chart(sheet.charts[chart]))


@api.route('/apps/<pid>/books/<book_name_or_ix>/sheets/<sheet_name_or_ix>/shapes', methods=['GET'])
@api.route('/books/<book_name_or_ix>/sheets/<sheet_name_or_ix>/shapes', methods=['GET'])
@api.route('/book/<path:fullname_or_name>/sheets/<sheet_name_or_ix>/shapes', methods=['GET'])
def get_shapes(sheet_name_or_ix, book_name_or_ix=None, fullname_or_name=None, pid=None):
    book = get_book_object(fullname_or_name, book_name_or_ix, pid)
    sheet = get_sheet_object(book, sheet_name_or_ix)
    return jsonify(shapes=[serialize_shape(shp) for shp in sheet.shapes])


@api.route('/apps/<pid>/books/<book_name_or_ix>/sheets/<sheet_name_or_ix>/shapes/<shape_name_or_ix>', methods=['GET'])
@api.route('/books/<book_name_or_ix>/sheets/<sheet_name_or_ix>/shapes/<shape_name_or_ix>', methods=['GET'])
@api.route('/book/<path:fullname_or_name>/sheets/<sheet_name_or_ix>/shapes/<shape_name_or_ix>', methods=['GET'])
def get_shape(sheet_name_or_ix, shape_name_or_ix, book_name_or_ix=None, fullname_or_name=None, pid=None):
    book = get_book_object(fullname_or_name, book_name_or_ix, pid)
    sheet = get_sheet_object(book, sheet_name_or_ix)
    shape = int(shape_name_or_ix) if shape_name_or_ix.isdigit() else shape_name_or_ix
    return jsonify(serialize_shape(sheet.shapes[shape]))


@api.route('/apps/<pid>/books/<book_name_or_ix>/sheets/<sheet_name_or_ix>/pictures', methods=['GET'])
@api.route('/books/<book_name_or_ix>/sheets/<sheet_name_or_ix>/pictures', methods=['GET'])
@api.route('/book/<path:fullname_or_name>/sheets/<sheet_name_or_ix>/pictures', methods=['GET'])
def get_pictures(sheet_name_or_ix, book_name_or_ix=None, fullname_or_name=None, pid=None):
    book = get_book_object(fullname_or_name, book_name_or_ix, pid)
    sheet = get_sheet_object(book, sheet_name_or_ix)
    return jsonify(pictures=[serialize_picture(pic) for pic in sheet.pictures])


@api.route('/apps/<pid>/books/<book_name_or_ix>/sheets/<sheet_name_or_ix>/pictures/<picture_name_or_ix>', methods=['GET'])
@api.route('/books/<book_name_or_ix>/sheets/<sheet_name_or_ix>/pictures/<picture_name_or_ix>', methods=['GET'])
@api.route('/book/<path:fullname_or_name>/sheets/<sheet_name_or_ix>/pictures/<picture_name_or_ix>', methods=['GET'])
def get_picture(sheet_name_or_ix, picture_name_or_ix, book_name_or_ix=None, fullname_or_name=None, pid=None):
    book = get_book_object(fullname_or_name, book_name_or_ix, pid)
    sheet = get_sheet_object(book, sheet_name_or_ix)
    pic = int(picture_name_or_ix) if picture_name_or_ix.isdigit() else picture_name_or_ix
    return jsonify(serialize_picture(sheet.pictures[pic]))


def run(host=None,
        port=None,
        debug=None,
        **options):
    """
    Run Flask development server
    """
    api.run(host=host, port=port, debug=debug,
            **options)


if __name__ == '__main__':
    run(debug=True)



from .def_list import plugin_def_list
from .extra import plugin_strikethrough, plugin_url
from .footnotes import plugin_footnotes
from .table import plugin_table
from .task_lists import plugin_task_lists

PLUGINS = {
    "url": plugin_url,
    "strikethrough": plugin_strikethrough,
    "footnotes": plugin_footnotes,
    "table": plugin_table,
    "task_lists": plugin_task_lists,
    "def_list": plugin_def_list,
}

__all__ = [
    "PLUGINS",
    "plugin_url",
    "plugin_strikethrough",
    "plugin_footnotes",
    "plugin_table",
]



import re
from ..inline_parser import LINK_LABEL
from ..scanner import unikey

__all__ = ['plugin_footnotes']

#: inline footnote syntax looks like::
#:
#:    [^key]
INLINE_FOOTNOTE_PATTERN = r'\[\^(' + LINK_LABEL + r')\]'

#: define a footnote item like::
#:
#:    [^key]: paragraph text to describe the note
DEF_FOOTNOTE = re.compile(
    r'( {0,3})\[\^(' + LINK_LABEL + r')\]:[ \t]*('
    r'[^\n]*\n+'
    r'(?:\1 {1,3}(?! )[^\n]*\n+)*'
    r')'
)


def parse_inline_footnote(inline, m, state):
    key = unikey(m.group(1))
    def_footnotes = state.get('def_footnotes')
    if not def_footnotes or key not in def_footnotes:
        return 'text', m.group(0)

    index = state.get('footnote_index', 0)
    index += 1
    state['footnote_index'] = index
    state['footnotes'].append(key)
    return 'footnote_ref', key, index


def parse_def_footnote(block, m, state):
    key = unikey(m.group(2))
    if key not in state['def_footnotes']:
        state['def_footnotes'][key] = m.group(3)


def parse_footnote_item(block, k, i, state):
    def_footnotes = state['def_footnotes']
    text = def_footnotes[k]

    stripped_text = text.strip()
    if '\n' not in stripped_text:
        children = [{'type': 'paragraph', 'text': stripped_text}]
    else:
        lines = text.splitlines()
        for second_line in lines[1:]:
            if second_line:
                break

        spaces = len(second_line) - len(second_line.lstrip())
        pattern = re.compile(r'^ {' + str(spaces) + r',}', flags=re.M)
        text = pattern.sub('', text)
        children = block.parse_text(text, state)
        if not isinstance(children, list):
            children = [children]

    return {
        'type': 'footnote_item',
        'children': children,
        'params': (k, i)
    }


def md_footnotes_hook(md, result, state):
    footnotes = state.get('footnotes')
    if not footnotes:
        return result

    children = [
        parse_footnote_item(md.block, k, i + 1, state)
        for i, k in enumerate(footnotes)
    ]
    tokens = [{'type': 'footnotes', 'children': children}]
    output = md.block.render(tokens, md.inline, state)
    return result + output


def render_ast_footnote_ref(key, index):
    return {'type': 'footnote_ref', 'key': key, 'index': index}


def render_ast_footnote_item(children, key, index):
    return {
        'type': 'footnote_item',
        'children': children,
        'key': key,
        'index': index,
    }


def render_html_footnote_ref(key, index):
    i = str(index)
    html = '<sup class="footnote-ref" id="fnref-' + i + '">'
    return html + '<a href="#fn-' + i + '">' + i + '</a></sup>'


def render_html_footnotes(text):
    return (
        '<section class="footnotes">\n<ol>\n'
        + text +
        '</ol>\n</section>\n'
    )


def render_html_footnote_item(text, key, index):
    i = str(index)
    back = '<a href="#fnref-' + i + '" class="footnote">&#8617;</a>'

    text = text.rstrip()
    if text.endswith('</p>'):
        text = text[:-4] + back + '</p>'
    else:
        text = text + back
    return '<li id="fn-' + i + '">' + text + '</li>\n'


def plugin_footnotes(md):
    md.inline.register_rule(
        'footnote',
        INLINE_FOOTNOTE_PATTERN,
        parse_inline_footnote
    )
    index = md.inline.rules.index('std_link')
    if index != -1:
        md.inline.rules.insert(index, 'footnote')
    else:
        md.inline.rules.append('footnote')

    md.block.register_rule('def_footnote', DEF_FOOTNOTE, parse_def_footnote)
    index = md.block.rules.index('def_link')
    if index != -1:
        md.block.rules.insert(index, 'def_footnote')
    else:
        md.block.rules.append('def_footnote')

    if md.renderer.NAME == 'html':
        md.renderer.register('footnote_ref', render_html_footnote_ref)
        md.renderer.register('footnote_item', render_html_footnote_item)
        md.renderer.register('footnotes', render_html_footnotes)
    elif md.renderer.NAME == 'ast':
        md.renderer.register('footnote_ref', render_ast_footnote_ref)
        md.renderer.register('footnote_item', render_ast_footnote_item)

    md.after_render_hooks.append(md_footnotes_hook)
